###################################################################################################
#
#  Ceetron Export
#  VTFxMinimal
#  
#  --------------------------------------------------------------------------------------------
#  Copyright (C) 2020, Ceetron AS
#  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
#  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
#  without the prior written permission of Ceetron AS.
###################################################################################################

import sys
import os


#--------------------------------------------------------------------------------------------------
# Set OS PATH and Python module search path
#--------------------------------------------------------------------------------------------------

thisDir = os.path.dirname(os.path.abspath(__file__))
myLibDir = os.path.abspath(thisDir + "/../../../lib/Release")

# Append the lib to the Python module search path
sys.path.append(myLibDir)
print("Appended directory to Python module search path: " + myLibDir)

# PATH environment variable only on Windows
if os.name == 'nt':
    currentPath = os.environ["PATH"] if "PATH" in os.environ else ""
    os.environ["PATH"] = myLibDir + ";" + currentPath
    print("Added directory to OS path: " + myLibDir)


#--------------------------------------------------------------------------------------------------
# Import required Export Python modules. Needs to be done after paths and sidedir
#--------------------------------------------------------------------------------------------------
from cee.core import *
from cee.vtfx import *

#-------------------------------------------------------------------------
# Initialize necessary components
#-------------------------------------------------------------------------
licenseKeyA = 0x00      # Todo: Set by developer 
licenseKeyB = 0x00      # Todo: Set by developer 

Globals.CEE_CORE_INSTANCE = CoreComponent.initialize(licenseKeyA, licenseKeyB)
VTFxComponent.initialize(Globals.CEE_CORE_INSTANCE)

#-------------------------------------------------------------------------
# Create the VTFx file
#-------------------------------------------------------------------------
file = File()

# Create VTFx file settings struct
fileSettings = FileSettings()
fileSettings.applicationName = Str("My first python app")
fileSettings.vendorName = Str("My Company")

# Let the API create the VTFx file 
fileName = Str("ExampleMinimal.vtfx");
file.create(fileName, fileSettings)

# Create a single database (could have many)
database = Database(file, Str("Only Database"), 1)    # database id 1

# Create a single case (could have many)
case = Case(file, Str("Only Case"), 1, 1)    # case id 1 using database with id 1

#-------------------------------------------------------------------------
# Add nodes
#-------------------------------------------------------------------------

# Create the node block
nodeBlock = NodeBlock(1, False)    # node block id 1, not using node ids

# Set the node data (Note: Ceetron Export expects interleaved node coordinates (xyzxyzxyz...))
nodeArr = [0.0, 0.0, 0.0,
           1.0, 0.0, 0.0,
           1.0, 1.0, 0.0,
           0.0, 1.0, 0.0,
           0.0, 0.0, 1.0,
           1.0, 0.0, 1.0,
           1.0, 1.0, 1.0,
           0.0, 1.0, 1.0]
nodeBlock.setNodes(nodeArr)

# Write the block
database.writeBlock(nodeBlock)

#-------------------------------------------------------------------------
# Add elements
#-------------------------------------------------------------------------

# Create the element block
elementBlock = ElementBlock(1, False, False)    # element block id 1, not using element ids, referring nodes by index
elementBlock.setNodeBlockId(1)    # Use node block with id = 1

# Set the element data
connectivities = [0,1,2,3,4,5,6,7]
elementBlock.addElements(ElementBlock.HEXAHEDRONS, connectivities)

# Write the element block
database.writeBlock(elementBlock)

#-------------------------------------------------------------------------
# Add geometry
#-------------------------------------------------------------------------

# Create geometry block
geometryBlock = GeometryBlock(1)    # Only one geometry per state

geoIndex = 0
partId = 1

# Set to use element block with id 1
geometryBlock.addElementBlock(geoIndex, 1, partId)    # Element block with id = 1

# Write the geometry block
database.writeBlock(geometryBlock)

# Create geometry info block
geometryInfoBlock = GeometryInfoBlock(1)    # Only one geometry per state
geometryInfoBlock.addPartInfo(geoIndex, partId, Str("Only Part"))

# Write the geometry info block
database.writeBlock(geometryInfoBlock)

#-------------------------------------------------------------------------
# Add one state
#-------------------------------------------------------------------------

# Create state info block
stateInfoBlock = StateInfoBlock()
stateInfoBlock.addStateInfo(1, Str("Only Step"), 42.0, StateInfoBlock.TIME)

# Write the state info block
database.writeBlock(stateInfoBlock)

#-------------------------------------------------------------------------
# Add optional properties
#-------------------------------------------------------------------------

# Set the part color to red
props = PropertySetCollection()
color = Color3f(1,0,0)

part_settings = PropertySet("part_settings")
part_settings.setValueUInt("context_geometry_index", geoIndex)
part_settings.setValueInt("context_part_id", partId)
part_settings.setValueColor3f("color", color)

props.addPropertySet(part_settings)
case.setProperties(props)


# Finally close the file
file.close()

print(Str("Exported successfully to file: ") + fileName)
