//##################################################################################################
//
//  Ceetron Export
//  VTFxQtSendToCloud
//  
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2017, Ceetron AS
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Ceetron AS. The contents of this file may 
//  not be disclosed to third parties, copied or duplicated in any form, in whole or in part,
//  without the prior written permission of Ceetron AS.
//##################################################################################################


#include <QCoreApplication>
#include <QObject>
#include <QtCore>
#include <QHttpMultiPart>
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QSslSocket>
#include <QStringList>
#include <QFile>

#include "CeeCore/CoreComponent.h"
#include "CeeCore/LogDestinationConsole.h"
#include "CeeCore/LogManager.h"
#include "CeeCore/Str.h"

#include "CeeVTFx/Database.h"
#include "CeeVTFx/GeometryBlock.h"
#include "CeeVTFx/File.h"
#include "CeeVTFx/StateInfoBlock.h"
#include "CeeVTFx/GeometryInfoBlock.h"
#include "CeeVTFx/NodeBlock.h"
#include "CeeVTFx/ElementBlock.h"
#include "CeeVTFx/VTFxComponent.h"

#include "QSCCloudTracker.h"

#include <locale.h>
#include <iostream>

#include "ceeqtUtilsCore.h"

QString cloudURL = "https://cloud.ceetron.com";

cee::PtrRef<cee::Instance>  g_componentInstance;

//--------------------------------------------------------------------------
// Create the VTFx file
// -------------------------------------------------------------------------
cee::Str createVTFxFile()
{
    cee::PtrRef<cee::vtfx::File> file = new cee::vtfx::File;

    // Create VTFx file settings struct
    cee::vtfx::FileSettings fileSettings;
    fileSettings.applicationName = "Ceetron Export: VTFxMinimal";
    fileSettings.vendorName = "My Company AS";

    // Let the API create the VTFx file 
    const cee::Str fileName = "ExampleMinimal.vtfx";
    if (!file->create(fileName, fileSettings))
    {
        std::cout << "#ERROR: Could not create file." << std::endl;
        return "";
    }

    // Create a single database (could have many)
    cee::PtrRef<cee::vtfx::Database> database = new cee::vtfx::Database(file.get(), "Only Database", 1); // database id 1

    // Create a single case (could have many)
    cee::PtrRef<cee::vtfx::Case> singleCase = new cee::vtfx::Case(file.get(), "Only Case", 1, 1); // case id 1 using database with id 1

    //--------------------------------------------------------------------------
    // Add nodes
    // -------------------------------------------------------------------------

    // Create the node block
    cee::PtrRef<cee::vtfx::NodeBlock> nodeBlock = new cee::vtfx::NodeBlock(1, false); // node block id 1, not using node ids

    // Set the node data (Note: Ceetron Export expects interleaved node coordinates (xyzxyzxyz...))
    const float NODES_PART[] =
    {
        0.0f, 0.0f, 0.0f,
        1.0f, 0.0f, 0.0f,
        1.0f, 1.0f, 0.0f,
        0.0f, 1.0f, 0.0f,
        0.0f, 0.0f, 1.0f,
        1.0f, 0.0f, 1.0f,
        1.0f, 1.0f, 1.0f,
        0.0f, 1.0f, 1.0f
    };

    std::vector<float> nodesPart(NODES_PART, NODES_PART + sizeof(NODES_PART) / sizeof(NODES_PART[0]));
    if (!nodeBlock->setNodes(nodesPart))
    {
        std::cout << "#ERROR: Error setting node block." << std::endl;
        return "";
    }

    // Write the block
    if (!database->writeBlock(nodeBlock.get()))
    {
        std::cout << "#ERROR: Error writing node block." << std::endl;
        return "";
    }


    //--------------------------------------------------------------------------
    // Add elements
    // -------------------------------------------------------------------------

    // Create the element block
    cee::PtrRef<cee::vtfx::ElementBlock> elementBlock = new cee::vtfx::ElementBlock(1, false, false); // element block id 1, not using element ids, referring nodes by index
    elementBlock->setNodeBlockId(1); // Use node block with id = 1

    // Set the element data
    const int CONNECTS_PART[] =
    {
        0, 1, 2, 3, 4, 5, 6, 7
    };

    std::vector<int> elementNodes(CONNECTS_PART, CONNECTS_PART + sizeof(CONNECTS_PART) / sizeof(CONNECTS_PART[0]));
    if (!elementBlock->addElements(cee::vtfx::ElementBlock::HEXAHEDRONS, elementNodes))
    {
        std::cout << "#ERROR: Error adding elements." << std::endl;
        return "";
    }

    // Write the element block
    if (!database->writeBlock(elementBlock.get()))
    {
        std::cout << "#ERROR: Error writing element block." << std::endl;
        return "";
    }


    //--------------------------------------------------------------------------
    // Add geometry
    // -------------------------------------------------------------------------

    // Create geometry block
    cee::PtrRef<cee::vtfx::GeometryBlock> geoBlock = new cee::vtfx::GeometryBlock(1); // Only one geometry per state

    size_t geoIndex = 0;
    int partId = 1;

    // Set to use element block with id 1
    if (!geoBlock->addElementBlock(geoIndex, 1, partId)) // Element block with id = 1
    {
        std::cout << "#ERROR: Error adding element block." << std::endl;
        return "";
    }

    // Write the geometry block
    if (!database->writeBlock(geoBlock.get()))
    {
        std::cout << "#ERROR: Error writing geometry block." << std::endl;
        return "";
    }

    // Create geometry info block
    cee::PtrRef<cee::vtfx::GeometryInfoBlock> infoBlock = new cee::vtfx::GeometryInfoBlock(1); // Only one geometry per state
    infoBlock->addPartInfo(geoIndex, partId, cee::Str("Only part"));

    // Write the geometry info block
    if (!database->writeBlock(infoBlock.get()))
    {
        std::cout << "#ERROR: Error writing info block." << std::endl;
        return "";
    }


    //--------------------------------------------------------------------------
    // Add one state
    // -------------------------------------------------------------------------

    // Create state info block
    cee::PtrRef<cee::vtfx::StateInfoBlock> stepInfo = new cee::vtfx::StateInfoBlock;
    if (!stepInfo->addStateInfo(1, "Only step", 42.0f, cee::vtfx::StateInfoBlock::TIME))
    {
        std::cout << "#ERROR: Error adding state info." << std::endl;
        return "";
    }

    // Write the state info block
    if (!database->writeBlock(stepInfo.get()))
    {
        std::cout << "#ERROR: Error writing state info block." << std::endl;
        return "";
    }

    // Finally close the file
    if (!file->close())
    {
        std::cout << "#ERROR: Error closing file." << std::endl;
        return "";
    }

    std::cout << "#INFO: Exported successfully to file: " << fileName.toStdString() << std::endl;

    return fileName;
}

//--------------------------------------------------------------------------------------------------
// SendToCloud app
// 
// Usage:
// sendToCloud -uploadId MY_UPLOAD_ID 
//--------------------------------------------------------------------------------------------------
int main(int argc, char *argv[])
{
    // On Linux, Qt will use the system locale, force number formatting settings back to "C" locale
    setlocale(LC_NUMERIC,"C");

    QCoreApplication app(argc, argv);
    CloudUploadTracker* uploadTracker = new CloudUploadTracker(&app, cloudURL);

    // Initialize components and set up logging to console
    // -------------------------------------------------------------------------
    unsigned int licenseKeyA = 0x00;    // Todo: Set by developer 
    unsigned int licenseKeyB = 0x00;    // Todo: Set by developer
    g_componentInstance = cee::CoreComponent::initialize(licenseKeyA, licenseKeyB);
    cee::vtfx::VTFxComponent::initialize(g_componentInstance.get());

    cee::PtrRef<cee::LogDestinationConsole> log = new cee::LogDestinationConsole();
    cee::CoreComponent::logManager()->setDestination(log.get());
    cee::CoreComponent::logManager()->setLevel("", 3);               // Log errors, warnings and info

#ifndef QT_NO_OPENSSL
    // Check that we have SSL/HTTPS support through Qt (normally provided via OpenSSL as a dynamic library)
    if (!QSslSocket::supportsSsl())
    {
        std::cerr << "Your system and/or Qt version does not support secure connections (HTTPS).\nAre the OpenSSL dynamic libraries missing?" << std::endl;
        return EXIT_FAILURE;
    }
#else
    // There is no compile time SSL support in this Qt version
    // Either download another Qt version or recompile Qt with OpenSSl support
    std::cerr << "Your Qt version was not built with OpenSSL support." << std::endl;
    return EXIT_FAILURE;
#endif

    QString uploadId = "";

    // Parse command line parameters
    QStringList arguments = app.arguments();

    if (arguments.size() == 1)
    {
        std::cout << "sendToCloud v. 1.0-0" << std::endl << "---------------------" << std::endl << std::endl;
        std::cout << "Note: App will upload to: " << cloudURL.toStdString() << std::endl << std::endl;
        std::cout << "Usage:" << std::endl << std::endl;
        std::cout << "> sendToCloud -uploadId MY_UPLOAD_ID" << std::endl << std::endl;
        return EXIT_FAILURE;
    }

    int uploadIdIndex = -1;
    for (int i = 1; i < arguments.size(); ++i)
    {
        if (uploadIdIndex == i) 
        {
            uploadId = arguments[i];
            uploadIdIndex = -1;
        } 
        else if (arguments[i].toLower() == "-uploadid")
        {
            uploadIdIndex = i + 1;
        }
    }

    // Check that all the needed info has been provided
    if (uploadId.isEmpty()) 
    {
        std::cout << "No Upload ID provided (-uploadId)" << std::endl;
        return EXIT_FAILURE;
    }
    
    cee::Str filename = createVTFxFile();

    // We're ready to SendToCloud!
    QHttpMultiPart *multiPart = new QHttpMultiPart(QHttpMultiPart::FormDataType);
    multiPart->setBoundary("boundary_Ceetron89906a4aJn1kvlAG52SFG2AD");

    QHttpPart vtfxPart;
    vtfxPart.setHeader(QNetworkRequest::ContentTypeHeader, QVariant("application/octet-stream"));
    vtfxPart.setHeader(QNetworkRequest::ContentDispositionHeader, QVariant("form-data; name=\"vtfx\" filename=\"upload.vtfx\""));

    QFile *file = new QFile(cee::qt::UtilsCore::toQString(filename));
    if (!file->open(QIODevice::ReadOnly))
    {
        std::cout << "Error opening file: " << filename.toStdString() << std::endl;
        return EXIT_FAILURE;
    }

    vtfxPart.setBodyDevice(file);

    file->setParent(multiPart); 
    multiPart->append(vtfxPart);

    QString uploadApp = "Ceetron Export Cloud Example";
    QString caseName = "Minimal example";
    QString urlString = QString(cloudURL + "/api/v1/models?uploadId=%1&uploadApp=%2&name=%3").arg(QString(QUrl::toPercentEncoding(uploadId))).arg(QString(QUrl::toPercentEncoding(uploadApp))).arg(QString(QUrl::toPercentEncoding(caseName)));
    std::cout << "#INFO: Using URL: " << urlString.toStdString() << std::endl;

    QUrl url(urlString);
    QNetworkRequest request(url);

    QNetworkAccessManager* manager = new QNetworkAccessManager(NULL);
    QNetworkReply* reply = manager->post(request, multiPart);
    multiPart->setParent(reply);

    QObject::connect(reply, SIGNAL(finished()), uploadTracker, SLOT(cloudUploadFinished()));
    QObject::connect(reply, SIGNAL(uploadProgress(qint64, qint64)), uploadTracker, SLOT(updateUploadProgress(qint64, qint64)));
    QObject::connect(uploadTracker, SIGNAL(finished()), &app, SLOT(quit()));

    std::cout << "#INFO: Uploading file to Ceetron Cloud..." << std::endl;

    int result = app.exec();
    std::cout << "#INFO: Done." << std::endl;

    return result;
}
